from typing import List
from .collection import Collection
from .scope import Scope
from .trace import Trace
from .point import Point


class Envelope(Collection):
    def __init__(self,
                 scope: Scope = None,
                 points: List[Point] = [],
                 traces: List[Trace] = []) -> None:
        """
        Constructor for an Envelope. Represents a collection of points and
        traces where traces are also theoretically collections of points.

        In Mass Spectrometry (specifically ms 1), an Envelope is commonly
        referred to as a ms 1 feature.

        Where this feature signifies that some sort of compound is
        reacting in the sample. Based off of this feature's
        data, one can build isotopic patterns, retrieve the
        monoisotopic peak, and so on.

        Args:
            scope (Scope, optional): The m/z and Retention time bounds of this
            envelope. See Scope class for more information. Defaults to None.
            points (List[Point], optional): List of MS Points. See Point class
            for more information. Defaults to [].
            traces (List[Trace], optional): List of Traces associated with
            this envelope. See Trace class for more information. Defaults to []
        """
        super().__init__(scope=scope, points=points)
        self.traces = traces

    def add_traces(self, traces: List[Trace]) -> None:
        """
        Add traces to this envelope's current list of traces.

        Args:
            traces (List[Trace]): The traces to add.
        """
        for trace in traces:
            self.traces.append(trace)

    def charge_state(self) -> int:
        """
        Computes the charge state from this envelope's current set of traces.

        Returns:
            int: The value representing the charge state of this envelope.
        """
        pass

    def monoisotopic_peak(self) -> Point:
        """
        Determines the monoisotopic peak from this envelope's current
        set of traces.

        Returns:
            Point: A Point instance representing this envelope's
            monoisotopic peak
        """
        pass

    def top_down(self) -> None:
        """
        TODO
        """
        pass

    def isotopic_distribution(self) -> None:
        """
        TODO
        """
        pass



from typing import List, Tuple

from .point import Point


class Utils:
    def __init__(self) -> None:
        pass

    @staticmethod
    def to_mspoints(
            rawpoints: List[Tuple[float]] = [],
            sort_by_intensity_asc: bool = False,
            sort_by_intensity_desc: bool = False
         ) -> List[Point]:
        """
        Converts a list of tuples that contain floats into
        a list of Points.

        Args:
            rawpoints (List[Tuple[float]], optional): The list of raw data.
            Defaults to [].
            sort_by_intensity_asc (bool, optional): Whether or not to have the
            returned list of points sorted by increasing intensity.
            Defaults to False.
            sort_by_intensity_desc (bool, optional): Whether or not to have the
            returned list of points sorted by decreasing intensity.
            Defaults to False.

        Returns:
            List[Point]: the converted list of Points.
        """

        # TODO: We could have a better sorting/ordering system here

        points = []
        for pt in rawpoints:
            converted = Point(mz=pt[0], rt=pt[1], intensity=pt[2])
            points.append(converted)

        if sort_by_intensity_asc:
            return sorted(points, key=lambda pt: pt.intensity)
        if sort_by_intensity_desc:
            return sorted(points, key=lambda pt: pt.intensity, reverse=True)

        return points

    @staticmethod
    def to_ndarray(
            points: List[Point],
            attributes_to_include: List[str] = ['mz', 'rt', 'intensity']
         ) -> List[List[float]]:
        """
        Converts a list of Points to a two dimensional list of floats.

        Args:
            points (List[Point], optional): [description].
            Defaults to [].
            attributes_to_include (List[str], optional): [description].
            Defaults to ['mz', 'rt', 'intensity'].

        Returns:
            List[List[float]]: The converted ndarray.
        """
        ndarray = []
        for pt in points:
            pt_as_list = pt.to_list(attributes=attributes_to_include)
            ndarray.append(pt_as_list)

        return ndarray

    @staticmethod
    def pluck_dimension(
            points: List[Point],
            dimension: str = 'mz'
         ) -> List[float]:
        """
        Plucks a single dimension from a list of points.

        Args:
            points (List[Point]): The Points you wish to convert.
            dimension (str, optional): The dimension you wish to pluck.
            Can be one of 'mz', 'rt', or 'intensity'. Defaults to 'mz'.

        Returns:
            List[float]: List of float values representing
            the plucked dimension.
        """
        # TODO: Pretty naive, def a better way of doing this.
        plucked = []
        for pt in points:
            if dimension == 'mz':
                plucked.append(pt.mz)
            if dimension == 'rt':
                plucked.append(pt.rt)
            if dimension == 'intensity':
                plucked.append(pt.intensity)

        return plucked



from typing import List, Tuple
import uuid

from .scope import Scope
from .point import Point
from .math import ConvexHull
from .utils import Utils
from .viz.scatter import Scatter3D
from .viz.bar import Bar, Bar3D


class Collection:
    def __init__(self,
                 id: int = None,
                 scope: Scope = None,
                 points: List[Point] = []) -> None:
        """
        The constructor of the Collection class. Provides the ability to group
        together raw Mass Spec data points and have easy to access useful
        operations on this group of data points.

        Args:
            id (int, optional): Provide an id for the collection or the
            collection will generate one. Defaults to None.
            scope (Scope, optional): The m/z and RT limits of the collection,
            See Scope class for more details. Defaults to None.
            points (List, optional): A list of Point objects. Defaults to [].
        """
        self.id = id if id is not None else uuid.uuid4()
        self.scope = scope if scope is not None else Scope()
        self.points = points

    def avg_mz(self) -> float:
        """
        Computes the average m/z value from this collection's current list of
        points.

        Returns:
            float: The average m/z value
        """
        if len(self.points) < 1:
            return 0.0

        mz_sum = sum(pt.mz for pt in self.points)
        return (mz_sum / len(self.points))

    def avg_rt(self) -> float:
        """
        Computes the average Retention Time value from this collection's
        current list of points.n.

        Returns:
            float: The average Retention Time value.
        """
        if len(self.points) < 1:
            return 0.0

        rt_sum = sum(pt.rt for pt in self.points)
        return (rt_sum / len(self.points))

    def avg_intensity(self) -> float:
        """
        Computes the average Intensity value from this collection's current
        list of points.

        Returns:
            float: The average Intensity value.
        """
        if len(self.points) < 1:
            return 0.0

        int_sum = sum(pt.intensity for pt in self.points)
        return (int_sum / len(self.points))

    def mz_bounds(self) -> Tuple[float]:
        """
        Computes minimum and maximum m/z values from this collection's current
        list of points.

        Returns:
            Tuple[float]: The min and max m/z values e.g. [432.32, 455.43]
        """
        if len(self.points) < 1:
            return 0, 0

        min_mz = self.points[0].mz
        max_mz = self.points[0].mz

        for p in self.points:
            if p.mz < min_mz:
                min_mz = p.mz
            if p.mz > max_mz:
                max_mz = p.mz

        return min_mz, max_mz

    def rt_bounds(self) -> Tuple[float]:
        """
        Computes the minimum and maximum Retention Time values from this
        collection's current list of points.

        Returns:
            Tuple[float]: The min and max Retention Time values e.g.
            [1202.2090, 1394.3032]
        """
        if len(self.points) < 1:
            return 0, 0

        min_rt = self.points[0].rt
        max_rt = self.points[0].rt

        for p in self.points:
            if p.rt < min_rt:
                min_rt = p.rt
            if p.rt > max_rt:
                max_rt = p.rt

        return min_rt, max_rt

    def add_points(self, points: List[Point]) -> None:
        """
        Add a list of points to this collection's current list of points.

        Args:
            points (List[Point]): The list of points to add (must be instances
            of the Point Class).
        """
        for pt in points:
            self.points.append(pt)

    def most_intense_point(self) -> Point:
        """
        Computes the most intense point in this collection's current
        list of points.

        Returns:
            Point: The most intense point
        """
        if len(self.points) < 1:
            return None

        most_intense = self.points[0]
        for pt in self.points:
            if pt.intensity > most_intense.intensity:
                most_intense = pt

        return most_intense

    def cumulative_intensity(self) -> float:
        """
        Computes the cumulative intensity from this collection's
        current list of points.

        Returns:
            float: The cumulative intensity value
        """
        return sum(pt.intensity for pt in self.points)

    def convex_hull(self) -> ConvexHull:
        """
        Builds a ConvexHull instance from this collection's current list of
        points. See ConvexHull class for more details.

        Returns:
            ConvexHull: an instance of the ConvexHull class
        """
        points_as_ndarray = Utils.to_ndarray(
            points=self.points, attributes_to_include=['mz', 'rt'])
        return ConvexHull(points=points_as_ndarray)

    def point_belongs(self, point: Point) -> bool:
        """
        Determines if it is reasonable to assume that an arbitrary point
        belongs to this envelope based off of its m/z and Retention Time scope.

        Args:
            point (Point): The point in question

        Returns:
            bool: Whether or not the point belongs
        """
        if ((self.scope.min_mz <= point.mz <= self.scope.max_mz) and
                (self.scope.min_rt <= point.rt <= self.scope.max_rt)):
            return True
        return False

    def as_dataframe(self):
        # TODO
        pass

    def three_d(self):
        """
        TODO: Document
        """
        # TODO: Pretty unideal, should be able to have a method that
        # returns a numpy array and we can just perform array slicing on
        # the dimensions.
        mz_data = Utils.to_ndarray(
            points=self.points, attributes_to_include=['mz'])
        rt_data = Utils.to_ndarray(
            points=self.points, attributes_to_include=['rt'])
        int_data = Utils.to_ndarray(
            points=self.points, attributes_to_include=['intensity'])

        scatter3d = Scatter3D(x=mz_data, y=rt_data, z=int_data)
        scatter3d.plot()

    def spectrum(self):
        """
        TODO: Document
        """
        mz_data = Utils.pluck_dimension(points=self.points, dimension='mz')
        int_data = Utils.pluck_dimension(
            points=self.points, dimension='intensity')
        bar = Bar(x=mz_data, y=int_data)
        bar.plot()

    def three_d_spectrum(self):
        """
        TODO: Document
        """
        mz_data = Utils.pluck_dimension(points=self.points, dimension='mz')
        rt_data = Utils.pluck_dimension(points=self.points, dimension='rt')
        int_data = Utils.pluck_dimension(
            points=self.points, dimension='intensity')
        bar = Bar3D(x=mz_data, y=rt_data, z=int_data)
        bar.plot()



from typing import List, Tuple


class Scope:
    def __init__(self, range: List[float] = None):
        """
        Constructor for Scope Object

        Args:
            range (List[float], optional): Range to be converted into scope.
            Must be in the format: [minmz, maxmz, minrt, maxrt].
            Defaults to None.
        """
        bounds = self.config_mins_maxes(range)
        self.min_mz, self.max_mz, self.min_rt, self.max_rt = bounds

    def config_mins_maxes(self, range: List[float] = None) -> Tuple[float]:
        """
        Configures the Scope's m/z and Retention Time's minimum
        and maximum values.

        Args:
            range (List[float], optional): The values to use for configuration.
            Defaults to None.

        Returns:
            Tuple[float]: The min and max values configured from the
            specified range.
        """
        # TODO: This should be 'private' ??
        if range is None:
            return 0.0, 0.0, 0.0, 0.0

        return range[0], range[1], range[2], range[3]

    def as_string(self) -> str:
        """
        Returns the Scope's properties as a string.
        Tends to be helpful for debugging purposes.

        Returns:
            str: The Scope's properties as a string.
        """
        return f"mz_{self.min_mz}_{self.max_mz}_rt_{self.min_rt}_{self.max_rt}"



from typing import List
import matplotlib.pyplot as plt


class Bar:
    def __init__(self, x: List[float], y: List[float]) -> None:
        """
        Constructor for the bar class. Serves as a wrapper around
        matplotlib's bar charts.

        Args:
            x (List[float]): The values to be plotted with
            respect to the x-axis.
            y (List[float]): The values to be plotted with
            respect to the y-axis.
        """
        self.x = x
        self.y = y

    def plot(self) -> None:
        """
        Renders the bar chart using matplotlib api.

        """
        # TODO: Pretty barebones in terms of flexibility.
        # What else should go here? :,)
        plt.figure()
        plt.bar(self.x, self.y, width=0.1, color='black')
        plt.xlabel('m/z', fontweight='bold', color='black',
                   fontsize='13', horizontalalignment='center')
        plt.ylabel('Signal Intensity', fontweight='bold',
                   color='black', fontsize='13', horizontalalignment='center')
        plt.show()


class Bar3D:
    def __init__(self, x: List[float], y: List[float], z: List[float]) -> None:
        """
        Constructor for the Bar3D class. Serves as a wrapper
        around matplotlib's 3d bar projection.

        Args:
            x (List[float]): The values to be plotted with respect
            to the x-axis.
            y (List[float]): The values to be plotted with respect
            to the y-axis.
            z (List[float]): The values to be plotted with respect
            to the z-axis.
        """
        self.x = x
        self.y = y
        self.z = z

    def plot(self) -> None:
        """
        Renders the 3D bar chart projection.

        """

        # Pretty barebones in terms of flexibility
        # What else do we need here??

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        zpos = [0 for i in self.x]
        dx = [1 for i in self.x]
        dy = [1 for i in self.x]

        ax.bar3d(self.x, self.y, zpos, dx, dy, self.z, color='#FFADF5')

        ax.set_xlabel('m/z')
        ax.set_ylabel('Retention Time')
        ax.set_zlabel('Signal Intensity')
        plt.show()



from typing import List


class ConvexHull:
    def __init__(self, points: List[float] = []) -> None:
        """
        Constructor for the ConvexHull. Provides an interface for building
        the convex hull from a list of floats.

        The convex hull of a shape is the smallest convex set that contains it.
        See https://en.wikipedia.org/wiki/Convex_hull for more information on
        convex hull.

        Args:
                points (List[float], optional): [description]. Defaults to [].
        """
        # Sort the points lexicographically
        # (tuples are compared lexicographically).
        # The 'set' here is for removal of duplicates to detect the case
        # we have just one unique point.
        self.points = sorted(set(tuple(map(tuple, points))))

    def hull(self) -> List[float]:
        """
        Builds the convex hull.

        Returns:
                List[float]: List of floats representing the boundaries
                of the convex hull.
        """
        # Boring case: no points or a single point, possibly
        # repeated multiple times.
        if len(self.points) <= 1:
            return self.points

        # 2D cross product of OA and OB vectors, i.e. z-component of their
        # 3D cross product. Returns a positive value, if OAB makes a
        # counter-clockwise turn, negative for clockwise turn, and
        # zero if the points are collinear.
        def cross(o, a, b):
            return (
                (a[0] - o[0]) * (b[1] - o[1]) - (a[1] - o[1]) * (b[0] - o[0])
            )

        # Build lower hull
        lower = []
        for p in self.points:
            while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
                lower.pop()
            lower.append(p)

        # Build upper hull
        upper = []
        for p in reversed(self.points):
            while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
                upper.pop()
            upper.append(p)

        # Concatenation of the lower and upper hulls gives the convex hull.
        # Last point of each list is omitted because it is repeated at
        # the beginning of the other list.
        return lower[:-1] + upper[:-1]



from typing import List
import matplotlib.pyplot as plt


class Scatter:
    def __init__(self) -> None:
        pass

    def plot(self) -> None:
        pass


class Scatter3D:
    def __init__(self,
                 x: List[float],
                 y: List[float],
                 z: List[float],
                 ) -> None:
        """
        Constructor for Scatter3D. Basically serves as a wrapper for
        matplotlib's 3d scatter plot projection.

        Args:
            x (List[float]): Data that will be plotted with respect
            to the x-axis
            y (List[float]): Data that will be plotted with respect
            to the y-axis
            z (List[float]): Data that will be plotted with respect
            to the z-axis
        """
        self.x = x
        self.y = y
        self.z = z

    def plot(self) -> None:
        """
        Renders the 3D scatter plot projection.

        """
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        ax.scatter(self.x, self.y, self.z, c=self.z)

        ax.set_xlabel('m/z')
        ax.set_ylabel('Retention Time')
        ax.set_zlabel('Signal Intensity')
        plt.show()



from typing import List
import uuid


class Point:
    def __init__(self,
                 mz: float = 0.0,
                 rt: float = 0.0,
                 intensity: float = 0.0) -> None:
        """
        Constructor for the point class. Represents a single Mass Spec raw
        data point. Has primary attributes m/z, retention time, and intensity.

        Args:
            mz (float, optional): [description]. Defaults to 0.0.
            rt (float, optional): [description]. Defaults to 0.0.
            intensity (float, optional): [description]. Defaults to 0.0.
        """
        self.id = uuid.uuid4()
        self.mz = float(mz)
        self.rt = float(rt)
        self.intensity = float(intensity)

    def to_list(self,
                attributes: List[str] = ['mz', 'rt', 'intensity']
                ) -> List[float]:
        """
        Converts this point's current attributes to a list. This is especially
        helpful in porting to other utilities that primarily interface
        with ndarrays or things of a kind.

        Args:
            attributes (List[str], optional): The attributes you wish to see
            in the converted list. Defaults to ['mz', 'rt', 'intensity'].

        Returns:
            List[float]: The Point's attributes as a list.
        """
        point_as_list = []
        if 'mz' in attributes:
            point_as_list.append(self.mz)
        if 'rt' in attributes:
            point_as_list.append(self.rt)
        if 'intensity' in attributes:
            point_as_list.append(self.intensity)

        return point_as_list

    def as_string(self) -> str:
        """
        Returns a string representing the information of the point.
        Particularly useful for debugging purposes.

        Returns:
            str: The Points information as a string.
        """
        return (
            f"id: {self.id}, "
            f"mz: {self.mz}, "
            f"rt: {self.rt}, "
            f"int: {self.intensity}"
        )



from typing import List

from .collection import Collection
from .scope import Scope
from .point import Point


class Trace(Collection):
    def __init__(self, scope: Scope = None, points: List[Point] = []) -> None:
        """
        Constructor for a trace. Represents an isotopic trace
        which theoretically, is just a collection of ms1 raw points.

        Args:
            scope (Scope, optional): The m/z and Retention Time bounds.
            See Scope class for more information. Defaults to None.
            points (List[Point], optional): List of points associated with
            this trace. See Point class for more information,. Defaults to [].
        """
        super().__init__(scope=scope, points=points)

    def peak(self) -> Point:
        """
        Determines the peak of this isotopic trace.

        Returns:
            Point: Point instance representing the peak of this trace.
        """
        return self.most_intense_point()
